@@ -8,7 +8,7 @@ Released under GNU GPL version 3 or later
 Partly based on SDLog2Parser by Anton Babushkin
 '''
 
-import struct, time
+import struct, time, os
 from pymavlink import mavutil
 
 FORMAT_TO_STRUCT = {
@@ -71,6 +71,7 @@ class DFMessage(object):
                 self._d[name] = self._d[name].rstrip('\0')
             if mul is not None and apply_multiplier:
                 self._d[name] = self._d[name] * mul
+        self._fieldnames = fmt.columns
         self.__dict__.update(self._d)
 
     def get_type(self):
@@ -89,33 +90,65 @@ class DFReader(object):
     def __init__(self):
         # read the whole file into memory for simplicity
         self.msg_rate = {}
+        self.new_timestamps = False
+        self.px4_timestamps = False
+        self.px4_timebase = 0
+        self.timestamp = 0
         
     def _rewind(self):
         '''reset counters on rewind'''
         self.counts = {}
         self.counts_since_gps = {}
-        self.messages = {}
+        self.messages = { 'MAV' : self }
         self.flightmode = "UNKNOWN"
+        self.percent = 0
 
     def _gpsTimeToTime(self, week, sec):
         '''convert GPS week and TOW to a time in seconds since 1970'''
         epoch = 86400*(10*365 + (1980-1969)/4 + 1 + 6 - 2)
         return epoch + 86400*7*week + sec - 15
 
+    def _find_time_base_new(self, gps):
+        '''work out time basis for the log - new style'''
+        t = self._gpsTimeToTime(gps.Week, gps.TimeMS*0.001)
+        self.timebase = t - gps.T
+        self.new_timestamps = True
+
+    def _find_time_base_px4(self, gps):
+        '''work out time basis for the log - PX4 native'''
+        t = gps.GPSTime
+        self.timebase = (t - self.px4_timebase) * 1.0e-6
+        self.px4_timestamps = True
+
     def _find_time_base(self):
         '''work out time basis for the log'''
         self.timebase = 0
-        gps1 = self.recv_match(type='GPS', condition='GPS.Week!=0')
+        gps1 = self.recv_match(type='GPS', condition='getattr(GPS,"Week",0)!=0 or getattr(GPS,"GPSTime")!=0')
+        if gps1 is None:
+            self._rewind()
+            return
+
+        if 'GPSTime' in gps1._fieldnames:
+            self._find_time_base_px4(gps1)
+            return
+            
+        if 'T' in gps1._fieldnames:
+            # it is a new style flash log with full timestamps
+            self._find_time_base_new(gps1)
+            return
+        
         counts1 = self.counts.copy()
         gps2 = self.recv_match(type='GPS', condition='GPS.Week!=0')
         counts2 = self.counts.copy()
 
         if gps1 is None or gps2 is None:
+            self._rewind()
             return
         
         t1 = self._gpsTimeToTime(gps1.Week, gps1.TimeMS*0.001)
         t2 = self._gpsTimeToTime(gps2.Week, gps2.TimeMS*0.001)
         if t2 == t1:
+            self._rewind()
             return
         for type in counts2:
             self.msg_rate[type] = (counts2[type] - counts1[type]) / float(t2-t1)
@@ -125,6 +158,12 @@ class DFReader(object):
         
     def _adjust_time_base(self, m):
         '''adjust time base from GPS message'''
+        if self.new_timestamps:
+            return
+        if self.px4_timestamps:
+            return
+        if getattr(m, 'Week', None) is None:
+            return
         t = self._gpsTimeToTime(m.Week, m.TimeMS*0.001)
         deltat = t - self.timebase
         if deltat <= 0:
@@ -140,9 +179,20 @@ class DFReader(object):
 
     def _set_time(self, m):
         '''set time for a message'''
-        rate = self.msg_rate.get(m.fmt.name, 50.0)
-        count = self.counts_since_gps.get(m.fmt.name, 0)
-        m._timestamp = self.timebase + count/rate
+        if self.px4_timestamps:
+            m._timestamp = self.timebase + self.px4_timebase
+        elif self.new_timestamps:
+            if m.get_type() == 'GPS':
+                m._timestamp = self.timebase + m.T*0.001
+            elif 'TimeMS' in m._fieldnames:
+                m._timestamp = self.timebase + m.TimeMS*0.001
+            else:
+                m._timestamp = self.timestamp
+        else:
+            rate = self.msg_rate.get(m.fmt.name, 50.0)
+            count = self.counts_since_gps.get(m.fmt.name, 0)
+            m._timestamp = self.timebase + count/rate
+        self.timestamp = m._timestamp
 
     def recv_msg(self):
         return self._parse_next()
@@ -160,6 +210,8 @@ class DFReader(object):
         else:
             self.counts_since_gps[type] += 1
 
+        if type == 'TIME' and 'StartTime' in m._fieldnames:
+            self.px4_timebase = m.StartTime * 1.0e-6
         if type == 'GPS':
             self._adjust_time_base(m)
         if type == 'MODE':
@@ -246,6 +298,8 @@ class DFReader_binary(DFReader):
         self.remaining -= fmt.len-3
         m = DFMessage(fmt, elements, True)
         self._add_msg(m)
+
+        self.percent = 100.0 * (self.offset / float(len(self.data)))
         
         return m
 
@@ -290,6 +344,13 @@ class DFReader_text(DFReader):
             if len(elements) >= 2:
                 break
 
+        # cope with empty structures
+        if len(elements) == 5 and elements[-1] == ',':
+            elements[-1] = ''
+            elements.append('')
+
+        self.percent = 100.0 * (self.line / float(len(self.lines)))
+
         if self.line >= len(self.lines):
             return None
 
